#!/usr/bin/python2.2
#^--- change the above line to point to your installation of Python


# $Id: //guest/mitch_stuart/perforce/utils/delta/p4Delta.py#1 $


import sys

import p4Util


USAGE_MESSAGE = """
Usage:

  p4Delta.py <dirspec1> <unqual-filespec1> <dirspec2> <unqual-filespec2>

  p4Delta.py <dirspec> <unqual-filespec1> <unqual-filespec2>

  p4Delta.py -help
"""

HELP_MESSAGE = """
INTRODUCTION
p4Delta.py finds the delta between two filesets.

The "delta" is the list of files that have been added, removed, or
updated.

A "fileset" is a group of one or more files that can be described with
a Perforce directory/file specification.  Perforce wildcards and modifiers
(label names, dates, etc.) can be used to specify the fileset.

COMMAND LINE SYNTAX
p4Delta.py supports two command line formats:

(a) General syntax that finds the delta between any two filesets:

  p4Delta.py <dirspec1> <unqual-filespec1> <dirspec2> <unqual-filespec2>

(b) Shortcut syntax that finds the delta between any two points in time
   in the same branch or directory:

  p4Delta.py <dirspec> <unqual-filespec1> <unqual-filespec2>

The differences between syntax (a) and (b) are discussed following the
examples section below.

Sample Invocation:

  p4Delta.py //depot/dev/console/suzuki ...@suzuki-label-1 ...@suzuki-label-5

Sample Output:

  Removed Files:
   manager/web/processMBeanChanges.jsp
  
  Added Files:
   JConsole/CondenserXMLUtil.java
   JConsole/com/fineground/console/version/ConsoleVersion.java
   mbeans/src/fineground/FGMBeanMetaInfo.java
   reports/web/jsp/scheduledReports.jsp
  
  Updated Files:
   JConsole/CAppContextBase.java
   JConsole/DbUtil.java
   JConsole/IAppContext.java
   . . .


FILE AND DIRECTORY SPECS
File and directory specifications can be in any syntax supported by Perforce:
depot syntax, client syntax, or local filesystem syntax (see:
http://www.perforce.com/perforce/doc.022/manuals/cmdref/o.fspecs.html).

A <dirspec> is the name of a directory.  It may be absolute (with a full path
from the root) or relative to the current directory.  It must not end with
a file separator character (/ or \) and it must not include any filename
component.  Examples of valid dirspecs are:

   dir1                 [local syntax, relative]
   C:\dev\dir1          [local syntax, absolute]
   /home/jdoe/dev/dir1  [local syntax, absolute]
   //depot/dev/dir1     [Perforce depot syntax]
   //jdoe/dev/dir1      [Perforce client syntax]

An <unqual-filespec> is the name of a file.  It may include Perforce wildcards.
It may include a Perforce modifier such as a branch name, date, etc.  (See the
Perforce File Specification URL referenced above for details on modifiers.)
It must not include any directory component.  When the filespec includes
wildcards or modifiers, it often needs to be quoted to prevent the shell
from expanding the argument.  Examples of valid unqual-filespecs are:

   "myfile.c"           [specific file]
   "myfile.c@label1"    [specific file with label]
   "*"                  [all files in dir (but not subdirs)]
   "..."                [all files in dir and subdirs]
   "*@label1"           [all files in dir with label]
   "...@label1"         [all files with label]
   "...@2003/07/01"     [all files as of date]
   "...#have"           [all files current version in client workspace]
   "...#head"           [all files current version in depot - same as "..."]
   "..."                [same as "...#head"]


Below are several examples of each syntax; following the examples is a
discussion of when to use each syntax.


EXAMPLES
Delta between two labels:

  p4Delta.py //depot/dev/console/suzuki ...@suzuki-label-3 ...@suzuki-label-4

Delta between two labels (specific file):

  p4Delta.py //depot/dev/console/suzuki myfile.c@suzuki-label-3 myfile.c@suzuki-label-4

Delta between a label and the current depot contents:

  p4Delta.py //depot/dev/console/suzuki ...@suzuki-label-3 ...

Delta between a label and the current client contents:

  p4Delta.py //depot/dev/console/suzuki ...@suzuki-label-3 ...#have

Delta between a date and the current depot contents:

  p4Delta.py //depot/dev/console/suzuki ...@2003/07/01 ...

Delta between the current client contents and the current depot contents
(similar to p4 sync -n: shows what is out of date):

  p4Delta.py //depot/dev/console/suzuki ...#have ...

Delta between two branches:

  p4Delta.py //depot/dev/console/ducati ... //depot/dev/console/suzuki ...

Delta between two branches, with specific labels:

  p4Delta.py //depot/dev/console/ducati ...@ducati-build-3 //depot/dev/console/suzuki ...@suzuki-build-2


COMPARISON OF SYNTAX FORMS
Syntax (a) is the more general form - any command that you can run with
syntax (b) can be rewritten into syntax (b), but the reverse is not true:
when you are delta'ing different directories or branches, you MUST use
syntax (a).

Syntax (b) is simply provided for convenience, because in most cases the
filespec between the two delta points of interest is the same, it is only the
modifier (the branch name, date, etc.) that is different.

For example, this command in syntax (b):

  p4Delta.py //depot/dev/console/suzuki ...@suzuki-label-3 ...@suzuki-label-4

can be expressed as follows in syntax (a):

  p4Delta.py //depot/dev/console/suzuki ...@suzuki-label-3 //depot/dev/console/suzuki ...@suzuki-label-4

The results of the two commands are the same, but syntax (a) just requires
more typing.

ORDER OF ARGUMENTS
Just as with any "diff" type utility, the order that you specify the
arguments is important: generally, you want the "earlier" argument on the
left and the "later" argument on the right.  For example:

  p4Delta.py //depot/myprog ...@label1 ...@label2

will show the delta between label1 and label2.  If you instead run:

  p4Delta.py //depot/myprog ...@label2 ...@label1

then the sense of the delta will be reversed.  In other words, files that
were removed between label1 and label2 will be shown as having been
added between label2 and label1.

Labels usually have a known sequence that makes it easy to determine the
correct order for the arguments.  But for branches, when you have parallel
development, there is no automatic way to know which of the following
will produce the most useful output:

  p4Delta.py //depot/myprog/rel1 ... //depot/myprog/rel2 ...

   or

  p4Delta.py //depot/myprog/rel2 ... //depot/myprog/rel1 ...

Both commands will show you the delta between the branches, but depending
on which branch you are more familiar with, or which branch has been more
active, you may find one output is easier to work with than the other.
You can just try both commands and see which one you prefer.

RELATION OF ARGUMENTS
Just as with any "diff" type utility, running a delta only makes sense if
there is some relationship between the filespecs.  For example, if you run

  p4Delta.py //depot/dir1 ... //depot/dir2 ...

on two random directories, dir1 and dir2, that are not related by any
branching or other common ancestry, the output will not be useful.  It will
simply tell you that every file in dir1 has been "deleted" and every file
in dir2 has been "added".
"""


def displayFilesetDelta(dirspecs, filespecs):

   try:
      deltaResults = computeFilesetDelta(dirspecs, filespecs)
      emitResults(deltaResults)
   except p4Util.p4Exception, msg:
      print msg



def computeFilesetDelta(dirspecs, filespecs):
   for i in range(2):
      p4Util.checkDirectoryExists(dirspecs[i])

   # fullspecs[0] and [1] are the full directory/file specifications

   fullspecs = [None] * 2

   for i in range(2):
      fullspecs[i] = p4Util.joinPathAndFilename(dirspecs[i], filespecs[i])

   # files[0] and [1] are lists of dictionaries, where each entry in
   # the list corresponds to one result from the "p4 files" command.

   files = [None] * 2
   
   for i in range(2):
      files[i] = p4Util.execP4Marshal("files " + fullspecs[i])

   foundError = 0

   for i in range(2):

      # This is not an error - you could have a valid, empty directory/label
      # if len(files[i]) == 0:
      #    print "Error: no files found: %s" % fullspecs[i]
      #    return

      if files[i][0].has_key("code") and files[i][0]["code"] == "error":
         raise p4Util.p4Exception(files[i][0]["data"].rstrip())

   # rootDepotPaths[0] and [1] are the Perforce-depot-syntax paths
   # corresponding to dirspecs[0] and [1].  If the dirspecs are already
   # in Perforce depot syntax, then rootDepotPaths are the same as dirspecs.
   # If the dirspecs are in Perforce client syntax or local syntax, then
   # the rootDepotPaths are the Perforce-depot-syntax paths that map
   # to dirspecs.
   #
   # For example, say we have a directory //depot/dev/foo/bar that is
   # mapped to R:\dev\foo\bar.  Say we are in the R:\dev directory
   # and we run: p4Delta.py foo\bar ...@label1 ...@label2.  The rootDepotPath
   # will be //depot/dev/foo/bar.

   rootDepotPaths = [None] * 2

   for i in range(2):
      rootDepotPaths[i] = p4Util.getDepotPath(dirspecs[i])

   # filesByRelativeName[0] and [1] are dictionaries where the values are
   # the members of files[0] and [1].  The keys are the file path,
   # relative to the rootDepotPath computed above.
   #
   # For example, say we have a file //depot/dev/foo/bar/blat/myfile.c,
   # the rootDepotPath as in the above example is //depot/dev/foo/bar.
   # The key for filesByRelativeName will be "blat/myfile.c" - the portion of
   # the path relative to the staring directory, plus the filename.

   filesByRelativeName = [None] * 2

   for i in range(2):
      filesByRelativeName[i] = mapFilesByRelativeName(rootDepotPaths[i],
         files[i])

   deltaResults = computeDelta(filesByRelativeName)
   return deltaResults


def mapFilesByRelativeName(rootDepotPath, files):
   startPath = rootDepotPath + "/"
   relativeNameIndex = len(startPath)

   filesByRelativeName = {}

   for file in files:
      relativeName = file["depotFile"][relativeNameIndex:]
      filesByRelativeName[relativeName] = file

   return filesByRelativeName


def computeDelta(filesByRelativeName):

   # Remove any deleted files from the lists.

   for i in range(2):
      relativeNames = filesByRelativeName[i].keys()

      for relativeName in relativeNames:
         if filesByRelativeName[i][relativeName]["action"] == "delete":
            del filesByRelativeName[i][relativeName]

   # Keep track of relativeNames of files that have had different kinds
   # of changes (delete / add / update).

   deletedFiles = []
   addedFiles = []
   updatedFiles = []

   # dict1 is the "driver" - loop through and check if the file has
   # changed in dict2.

   dict1 = filesByRelativeName[0]
   dict2 = filesByRelativeName[1]

   relativeNames = dict1.keys()

   # Process all the files in dict1.

   for relativeName in relativeNames:

      # If the file is in dict1 but not in dict2, it has been deleted.

      if not dict2.has_key(relativeName):
         deletedFiles.append(relativeName)
         del dict1[relativeName]
         continue

      # See if the file has changed

      if hasFileChanged(dict1[relativeName], dict2[relativeName]):
         updatedFiles.append(relativeName)

      # Done with this file

      del dict1[relativeName]
      del dict2[relativeName]

   # There should not be any files left in dict1 - if there are, it is
   # a bug.

   if len(dict1) > 0:
      print "Error: the following files were not processed:"

      for relativeName in dict1.keys():
         print "  %s" % relativeName
   
   # Any files left in dict2 have been added.

   addedFiles[len(addedFiles):] = dict2.keys()

   return (deletedFiles, addedFiles, updatedFiles)


def hasFileChanged(fileInfo1, fileInfo2):

   # If the changelist number is the same, the file has not changed.

   if fileInfo1["change"] == fileInfo2["change"]:
      return 0

   # If both files have been edited, and the changelist number is
   # different, the file has changed.

   if fileInfo1["action"] == "edit" and fileInfo2["action"] == "edit":
      return 1

   # If we get here, the file has most likely been branched or integrated.
   # It may or may not have changed, so run p4 diff2 to check.

   diff2cmd = "diff2 -q " + \
      fileInfo1["depotFile"] + "#" + fileInfo1["rev"] + " " + \
      fileInfo2["depotFile"] + "#" + fileInfo2["rev"]

   diff2 = p4Util.execP4Marshal(diff2cmd)

   # If there are no differences, diff2 SHOULD return no output
   # (but see below for an exception).

   if len(diff2) == 0:
      return 0

   # Exception to expected diff2 output.
   # There is a bug in the p4 diff2 output.  It is supposed to return with
   # no output if the files are identical, instead it returns with an
   # error message (see Perforce tech support CALL # 542982).

   # Example of output from p4 -G diff2 -q
   #
   # No differences exist:
   #
   #   severity: 2
   #   data: No file(s) to diff.
   #   code: error
   #   generic: 17
   #
   # Differences exist:
   #
   #   code: info
   #   data: ==== //depot/dev/console/suzuki/build.sh#7 (text+klx) - //depot/dev/console/suzuki/build.sh#18 (text+klx) ==== content
   #   level: 0

   if diff2[0]["data"].startswith("No file(s) to diff"):
      return 0

   return 1


def emitResults(deltaResults):

   headings = ("Removed Files", "Added Files", "Updated Files")

   for i in range(len(deltaResults)):

      if len(deltaResults[i]) > 0:

         print "%s:" % headings[i]

         deltaResults[i].sort()
   
         for file in deltaResults[i]:
            print "  %s" % file
   
         print   # blank line between sections



def main():
   if len(sys.argv) > 1 and sys.argv[1] == "-help":
      print HELP_MESSAGE
      sys.exit(0)
   
   if len(sys.argv) < 4 or len(sys.argv) > 5:
      sys.exit(USAGE_MESSAGE)
      
   dirspecs = [None] * 2
   filespecs = [None] * 2
   
   dirspecs[0] = sys.argv[1]
   filespecs[0] = sys.argv[2]
   
   if len(sys.argv) == 5:
      dirspecs[1] = sys.argv[3]
      filespecs[1] = sys.argv[4]
   
   elif len(sys.argv) == 4:
      dirspecs[1] = sys.argv[1]
      filespecs[1] = sys.argv[3]
   
   displayFilesetDelta(dirspecs, filespecs)


if __name__ == '__main__':
   main()

